-- NOVE.EX
-- This is a simpler version of the analysis of variance for non-orthogonal
-- designs, it does calculate expected mean squares but it does not have
-- the option to transform the data, calculate regression lines nor does it
-- do difference effects. However, the input is relatively simple and the
-- program is quite fast.

constant e0=1E-11
constant TRUE=1, FALSE=0
constant dc = {"Sunday ","Monday ","Tuesday ","Wednesday ",
	    "Thursday ","Friday ","Saturday "}
constant mc = {"January ","February ",
	    "March ","April ","May ","June ","July ","August ",
	    "September ","October ","November ","December "}
constant AL="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
type boolean(integer x)
  return x=FALSE or x=TRUE
end type
include get.e
include math.e
include stat.e

-- Stat.e contains algorithms to calculate the F-distribution with
-- arbitrary real-valued, but positive numbers of degrees of freedom.

integer a,b,c,d,b1,c1,t,r,q,q1,s,s1,s2,r1,r2,h1,
    k1,k2,l1,b2,h,p,p1,i1,j1,i2,L,N,txt,out
boolean ts,n0,wt,dn
atom e1,w,z0,o,X,P,Q,u1,u2,u3,u4,u5
sequence  tx,ds,tm,vt,f,g,y,m,n,u,v,m1,M,J,S,Y,B,D,z,x,R,U
procedure nl(integer a)

-- This procedure does "a" new lines in a row

  for i=1 to a do puts(out,'\n')
  end for
end procedure
procedure gn()

-- This procedure generates the treatment codes in standard order when
-- it is indicated in the input that this is required, it is also used
-- to generate level codes during the output of means etc. towards the
-- end of the program run.

  integer a,b
  b=p if b>0 then
    a=f[b] while y[a]=g[a] do
      if b=1 then exit
      end if b=b-1 a=f[b]
    end while for j=b+1 to p do y[f[j]]=1
    end for y[a]=y[a]+1
  end if
end procedure
procedure ML()

-- This procedure calculates the vectors x and z for each treatment. These
-- vectors are used to build up the symmetric matrix S=X'X (s9 in this program,
-- used for the calculation of mean squares and effects) and D=X'Z (d9 in
-- this program, used for the calculation of expected mean squares).
-- Restrictions imposed on the estimates of the effects are built into the
-- matrices S and D and reflected in the vectors x, while z is the same vector
-- as x but without the restrictions built into it.
-- In addition to the calculation of expected mean squares the vector z is also
-- necessary for the generation of nested effects if applicable.

  sequence Z
  Z=repeat(0,p1) for i=1 to t do
    a=M[i][1] b=M[i][2] c=M[i][3] d=M[i][4]
    b1=b+1 if a=2 then 
      j1=n[c] k1=n[d] for j=1 to j1 do
	k2=k1*(j-1) o=x[u[c]+j]
	for k=1 to k1 do Z[k2+k]=rnd(o*x[u[d]+k])
	end for
      end for 
      n[b]=j1*k1 u[b1]=u[b]+n[b]
      for j=1 to n[b] do x[u[b]+j]=Z[j]
      end for j1=m[c] k1=m[d] for j=1 to j1 do
	k2=k1*(j-1) o=z[v[c]+j]
	for k=1 to k1 do Z[k2+k]=rnd(o*z[v[d]+k]) 
	end for
      end for 
      m[b]=j1*k1 v[b1]=v[b]+m[b]
      for j=1 to m[b] do z[v[b]+j]=Z[j] 
      end for elsif a=3 then  
      j1=n[c] k1=m[d] i1=m[c] for k=1 to k1 do
	k2=j1*(k-1) o=z[v[d]+k]
	for j=1 to j1 do Z[k2+j]=rnd(o*x[u[c]+j])
	end for
      end for
      n[b]=j1*k1 u[b1]=u[b]+n[b]
      for j=1 to n[b] do x[u[b]+j]=Z[j] 
      end for for k=1 to k1 do
	k2=i1*(k-1) o=z[v[d]+k]
	for j=1 to i1 do Z[k2+j]=rnd(o*z[v[c]+j])
	end for
      end for
      m[b]=i1*k1 v[b1]=v[b]+m[b]
      for j=1 to m[b] do z[v[b]+j]=Z[j]
      end for else 
      m[b]=g[c] n[b]=m[b]-1 u[b1]=u[b]+n[b] v[b1]=v[b]+m[b]
      for j=u[b]+1 to u[b1] do x[j]=0 
      end for if y[c]>g[c] then y[c]=g[c]
      end if
      for j=v[b]+1 to v[b1] do z[j]=0 
      end for
      if y[c]>=1 then
	if y[c]<=n[b] then x[u[b]+y[c]]=1 else
	  for j=u[b]+1 to u[b1] do x[j]=-1 
	  end for
	end if z[v[b]+y[c]]=1
      end if
    end if
  end for
end procedure
txt=open("input.txt","r") out=open("output.txt","w") dn=FALSE

-- Data are read from the file INPUT.TXT which can be prepared in advance and
-- the results of the analysis are written into the file OUTPUT.TXT in order
-- that the user can peruse the results before having them printed. In the 
-- case that different models are tried out this can save some paper.

puts(out,"Non-orthogonal ANOVA by Alexei van Baaren.") nl(1)
U=date() a=U[7] b=U[2] puts(out,dc[a]&mc[b]) printf(out,"%2d,",U[3])
printf(out,"%5d",1900+U[1]) nl(2)

-- At this point the descriptive text and the code numbers are being read
-- by the program which can then determine the size of the application.

tx=gets(txt) puts(out,tx) nl(1)
U=get(txt) s=rnd(U[2]) U=get(txt) l1=rnd(U[2]) U=get(txt) h1=0
t=rnd(U[2]) wt=FALSE if t<0 then wt=TRUE 
end if t=abs(t) h=t+2 u=repeat(0,h) v=repeat(0,h+1) if s>0 then dn=TRUE 
end if s=abs(s) s1=s+1 s2=s+2 n0=FALSE p1=0 k2=0 if l1=0 then n0=TRUE  
end if k1=0 c1=0 vt=repeat(0,h) M=repeat({},t) u[2]=1 v[2]=1
n=repeat(1,h) m=repeat(1,h) y=repeat(0,s) g=repeat(0,s) i2=0
ds=repeat({},s) M=repeat(repeat(0,4),t) m1=repeat(repeat(0,s2),h) 
f=repeat(0,h) f[2]=1 for i=1 to s do
  U=get(txt) g[i]=rnd(U[2]) ds[i]=gets(txt)
end for for i=1 to t do

-- At this point the size of the application is determined and preparations
-- are made. In languages which support dynamic array allocation like
-- ALGOL, QBASIC, EUPHORIA etc. the exact sizes of the arrays are being
-- determined.

  U=get(txt) a=rnd(U[2]) U=get(txt) b=rnd(U[2])+1 
  U=get(txt) c=rnd(U[2]) U=get(txt) d=rnd(U[2]) b1=b+1 if a=2 then
    c=c+1 d=d+1 n[b]=n[c]*n[d] m[b]=m[c]*m[d] 
    vt[b]=vt[c] or vt[d] for j=1 to s do m1[b][j]=m1[c][j]+m1[d][j] 
    end for elsif a=3 then
    c=c+1 d=d+1 j1=abs(m1[c][s1]) m1[b][s1]=sgn(j1)-1-j1
    n[b]=n[c]*m[d] m[b]=m[c]*m[d] vt[b]=TRUE
    for j=1 to s do m1[b][j]=m1[c][j]+2*sgn(m1[d][j])
    end for else
    m[b]=g[c] n[b]=m[b]-1 for j=1 to s do m1[b][j]=0
    end for y[c]=1 m1[b][c]=1 vt[b]=FALSE
  end if M[i]={a,b,c,d} u[b1]=u[b]+n[b] v[b1]=v[b]+m[b] 
  m1[b][s2]=i m1[b][s1]=u[b1]
  if v[b1]>k2 then k2=v[b1] 
  end if if m[b]>p1 then p1=m[b]
  end if  if u[b1]>k1 then k1=u[b1] 
  end if j1=m[b]*(m[b]+1)/2 f[b1]=f[b]+j1 if f[b1]>h1 then h1=f[b1]
  end if if j1>i2 then i2=j1
  end if
end for q1=M[t][2] r=q1+1 q=u[r] h=v[r] m1[r][s2]=t if k1>i2 then i2=k1
end if x=repeat(0,i2) 
z=repeat(0,k2) r1=r+1 r2=r1+1 tm=repeat({},r1) i2=q if p1>i2 then i2=p1
end if m1[r][s1]=q tm[1]="MEAN      " tm[r]="NON-FIT   " for i=2 to q1 do

-- Here the indicators of the modelterms are being put together using
-- letters of the alphabet, main effects are single letters, interactions
-- are combinations of letters and nested effects are also combinations
-- of letters of which some appear within parentheses, the latter are those
-- factors (also called dead factors) within which the factors outside the
-- parentheses (also called living factors) vary.

  j1=1 tm[i]="          " for j=1 to s do
    b=m1[i][j] a=abs(b) if a=1 then tm[i][j1]=AL[j] j1=j1+1
    end if
  end for if vt[i]=TRUE then
    tm[i][j1]='(' j1=j1+1 for j=1 to s do
      if m1[i][j]=2 then tm[i][j1]=AL[j] j1=j1+1
      end if
    end for tm[i][j1]=')'
  end if
end for tm[r1]="WITHIN    " for i=1 to s do
  y[i]=1 m1[r][i]=y[i] f[i]=i
end for y[s]=0 j1=floor((q+1)*(q+2)/2) N=0 L=0 m1[1][s1]=1

-- Initialization of some constants and arrays.

u1=0 u2=0 u3=0 u4=0 u5=0 x[1]=1 z[1]=1 p=s 
J=repeat(0,k1) for i=2 to k1 do i1=i-1 J[i]=J[i1]+i1
end for S=repeat(0,j1) Y=repeat(0,i2) B=repeat(0,h) D=repeat(repeat(0,h),q)
w=1 Q=0 P=0 n=repeat(1,r2) while TRUE do

-- From here on the actual data are being read and the matrices S and D
-- as well as the response-vectors B and Y built up.

  U=get(txt) if U[1]!=0 then exit 
  end if if dn then
    y[1]=rnd(U[2]) for i=2 to s do U=get(txt) y[i]=rnd(U[2])
    end for U=get(txt) else gn()

-- If s>0 read the levels of the next treatment otherwise generate them.

  end if if wt then

-- read a weight, not the same as number of replications, in de codes - t<0.

    w=U[2] U=get(txt)
  end if if n0 then l1=rnd(U[2]) 
  end if if l1>0 then

-- first read the data, then start building up the matrices S and D
-- as well as the response-vector. The vector Y has the restrictions on
-- the effects built into it, the vector B does not.

    if n0 then U=get(txt)
    end if z0=U[2] X=z0 Q=Q+w*z0*z0 
    for i=2 to l1 do
      U=get(txt) z0=U[2] X=X+z0 Q=Q+w*z0*z0
    end for P=P+w*X*X/l1 if w>e0 then
      ML() N=N+l1 L=L+1 o=w*l1 u1=u1+o u2=u2+o*l1  
      u3=u3+w u4=u4+w*w u5=u5+o*o for i=1 to q do
	z0=o*x[i] for j=i to q do
	  j1=i+J[j] S[j1]=S[j1]+z0*x[j]
	end for Y[i]=Y[i]+w*x[i]*X for j=1 to h do 
	  D[i][j]=D[i][j]+z0*z[j]
	end for
      end for for i=1 to h do
	B[i]=B[i]+w*z[i]*X
      end for
    end if
  end if 
end while

-- At this point the data have been read and the actual computations start.

close(txt) w=N/S[1] v[r1]=L for j=1 to h do
  z[j]=D[1][j] if z[j]>e0 then B[j]=B[j]/z[j] 
  end if
end for e1=0 for i=1 to q do

-- Determine a norm for the matrix S used during the inversion process.

  o=0 for j=1 to q do
    if j<i then i1=j+J[i] else i1=i+J[j] 
    end if o=o+abs(S[i1])
  end for if o>e1 then e1=o
  end if 
end for e1=w*e1 for l=1 to q1 do 

-- Here the inversion process starts.

  l1=l+1 for i=u[l]+1 to u[l1] do
    i1=i-1 for k=1 to i do x[k]=S[k+J[i]] 
    end for b2=i+1
    o=w*x[i] for k=1 to i1 do o=o-x[k]*x[k]
    end for if o/b2/e1>e0 then o=sqrt(o) else
      o=1E30 n[l]=n[l]-1
    end if S[i+J[i]]=o for j=b2 to q do
      j1=i+J[j] S[l]=w*S[l]
      for k=1 to i1 do S[j1]=S[j1]-x[k]*S[k+J[j]]
      end for if abs(S[j1])>e0 then S[j1]=S[j1]/o else S[j1]=0
      end if
    end for Y[i]=w*Y[i] for k=1 to i1 do
      Y[i]=Y[i]-x[k]*Y[k] 
    end for Y[i]=Y[i]/o for j=1 to h do
      D[i][j]=w*D[i][j] for k=1 to i1 do
	D[i][j]=D[i][j]-x[k]*D[k][j]
      end for D[i][j]=D[i][j]/o
    end for
  end for
end for k1=0 for i=1 to q1 do k1=k1+n[i] 
end for b1=N-k1 n[r]=L-k1 P=w*P n[r1]=N-L Q=w*Q 
for i=1 to s do puts(out,"   ") puts(out,AL[i]) 
  puts(out," = ") puts(out,ds[i]) 
end for nl(1) R=repeat(repeat(0,r1),r1) U=repeat(1,h1) e1=n[r]/(L-1) 

-- At this point the coefficients of the variance components for the
-- NON-FIT and WITHIN terms in the expected mean squares are calculated.
-- Actually these coefficients are approximations based on certain
-- assumptions. Calculating the exact contributions would take a lot of
-- memory and (in large applications) time. The coefficients are stored
-- in the array R. The approximations are fairly close anyway.

u1=w*u1 u2=w*u2 u3=w*u3 u4=w*u4 u5=w*u5 z0=u1/N for j=1 to q1 do
  o=n[j]/N R[j][r]=o*u2 R[j][r1]=o*u1*z0
end for R[r][r]=e1*(u1-u5/u1) R[r][r1]=e1*(u3-u4/u3)
R[r1][r1]=u1-u3 f[1]=0 f[2]=1 for j=1 to h do
  D[1][j]=z[j]
end for for l=1 to t do

-- At this point calculation starts of matrices with coefficients
-- of correlation between the effects. These coefficients are used
-- for the computation of the contributions of the modelterms to the
-- expected mean squares.

  a=M[l][1] b=M[l][2] c=M[l][3] d=M[l][4] if a=2 then
    for i=1 to m[c] do
      i2=m[d]*(i-1) i1=f[c]+i for g1=1 to m[d] do
	k1=i2+g1 b2=f[d]+g1 for j=i to m[c] do
	  l1=i1+J[j] j1=(j-1)*m[d] if j=i then
	    for k=g1 to m[d] do
	      k2=k+j1 x[k1+J[k2]]=U[b2+J[k]]
	    end for else for k=1 to m[d] do
	      k2=k+j1 if g1>k then c1=k+J[g1] else c1=g1+J[k]
	      end if x[k1+J[k2]]=U[l1]*U[f[d]+c1]
	    end for
	  end if
	end for
      end for
    end for m[b]=m[c]*m[d] for j=1 to m[b] do
      i1=f[b]+j for k=j to m[b] do
	U[i1+J[k]]=x[j+J[k]]
      end for
    end for elsif a=3 then 
    for j=1 to m[c] do
      b2=f[c]+j for k=j to m[c] do
	x[j+J[k]]=U[b2+J[k]]
      end for
    end for for j=1 to m[d] do
      i1=f[b]+m[c]*(j-1) for k=j to m[d] do
	i2=m[c]*(k-1) if j=k then
	  for i=1 to m[c] do
	    for g1=i to m[c] do
	      l1=i2+g1 U[i1+i+J[l1]]=x[i+J[g1]]
	    end for
	  end for else
	  for i=1 to m[c] do
	    for g1=1 to m[c] do
	      l1=i2+g1 U[i1+i+J[l1]]=0
	    end for
	  end for
	end if
      end for
    end for m[b]=m[c]*m[d] else
    j1=g[c] m[b]=j1 if d>=j1 then o=-1/(d-1) else o=0
    end if for j=1 to m[b] do
      i1=f[b]+j U[i1+J[j]]=1
      for k=j+1 to m[b] do
	U[i1+J[k]]=o
      end for
    end for
  end if f[b+1]=f[b]+m[b]*(m[b]+1)/2 
end for for i=2 to r1 do

-- Now the calculation of the expected mean squares is being completed, the
-- appropriate coefficients are also printed as a triangular matrix, the
-- columns represent the modelterms in reverse order. See the documentation
-- NOVGUIDE.TXT.

  if n[i]>0 then
    i1=i+1 for j=i to r1 do
      if j<r then
	j1=j+1 for k=u[i]+1 to u[i1] do
	  o=0 for l=v[j]+1 to v[j1] do
	    o=o+D[k][l]*D[k][l] k1=l-v[j]
	    i2=f[j]+k1 for g1=l+1 to v[j1] do
	      k2=g1-v[j] o=o+2*U[i2+J[k2]]*D[k][l]*D[k][g1]
	    end for
	  end for R[i][j]=R[i][j]+o
	end for
      end if R[i][j]=R[i][j]/n[i]
    end for
  end if
end for puts(out,"SOURCE       EXPECTED MEAN SQUARES") nl(1)
vt[r]=FALSE for i=2 to r1 do
  if n[i]>0 then
    puts(out,tm[i]) for j=r1 to i by -1 do
      printf(out,"%9.2f",R[i][j])
    end for nl(1)
  end if R[i][r]=1
end for nl(3) puts(out,"ANALYSIS OF VARIANCE")

-- Calculation of the analysis of variance table.

nl(2) if b1>0 then ts=TRUE 
  end if x[1]=Y[1]*Y[1] o=x[1] for i=2 to q1 do 
  x[i]=0 i1=i+1 for j=u[i]+1 to u[i1] do
    x[i]=x[i]+Y[j]*Y[j] 
  end for o=o+x[i]
end for U=repeat(0,2) x[r]=Q-o x[r1]=Q-P tm[r]="RESIDUAL  " n[r]=n[r]+n[r1]
puts(out,"SOURCE          SS           DF       MS")
if ts then puts(out,"           F   DF(den.)    P")
end if x[r2]=Q-x[1] nl(2) k1=N-1 for i=2 to r do
  if n[i]>0 then
    puts(out,tm[i]) z0=x[i]/n[i]
    printf(out,"%13.6f%8d%14.6f",{x[i],n[i],z0}) if ts then
      w=0 o=0 for j=i+1 to r do

-- At this point the program starts to calculate the test statistic
-- for the current term. The numerator is the term to be tested while the
-- denominator is a linear combination of all terms in the model following
-- the one to be tested. This linear combination is determined in such
-- a way that the coefficients of the variance components in the linear
-- combination are equal to the corresponding ones in the term to be
-- tested. This method is not quite correct because it is based on an
-- assumption which does not hold exactly, but in data collections where
-- non-orthogonality has not been designed there is not much choice.

	if n[j]>0 then
	  l1=n[j] e1=x[j]/l1 z[j]=R[i][j] for k=i+1 to j-1 do
	    z[j]=z[j]-z[k]*R[k][j]
	  end for z[j]=z[j]/R[j][j] e1=e1*z[j] o=o+e1 w=w+e1*e1/l1
	else z[j]=0
	end if
      end for if o>e0 then
	w=o*o/w e1=z0/o printf(out,"%9.2f%8.2f",{e1,w}) if w>0 then
	  o=n[i] e1=FD(o,w,e1) printf(out,"%9.3f",e1)
	end if
      end if
    end if nl(1)
  end if
end for nl(1) k1=N-1 N=N-L L=n[r]-n[r1] n[r]=L if N*L>0 then
  Q=Q-P P=x[r]-x[r1] X=P/L z0=Q/N
  printf(out,"  NON-FIT %13.6f%8d%14.6f",{P,L,X}) 
  o=X/z0 e1=FD(L,N,o) printf(out,"%9.2f        %9.3f\n",{o,e1})
  printf(out,"  WITHIN  %13.6f%8d%14.6f",{Q,N,z0}) nl(2)
end if
printf(out,"TOTAL     %13.6f%8d%14.6f\n",{x[r2],k1,x[r2]/k1})

-- This is the last stage where the effects are calculated and inversion
-- of the matrix S is completed.

p=0 l1=0 r1=2*s+9 for i=q to 1 by -1 do
  i1=i+J[i] o=S[i1] for j=i+1 to q do j1=i+J[j] Y[i]=Y[i]-Y[j]*S[j1]
  end for if o>e0 then Y[i]=Y[i]/o else Y[i]=0
  end if 
end for if b1>0 then for i=1 to q do
    k1=i+J[i] x[i]=1/S[k1] S[k1]=x[i] for j=i+1 to q do
      x[j]=0 j1=J[j] for k=i to j-1 do x[j]=x[j]-x[k]*S[k+j1]
      end for x[j]=x[j]/S[j+j1] S[i+j1]=x[j]
    end for 
    for j=1 to i do
      j1=j+J[i] S[j1]=x[i]*S[j1]
      for k=i+1 to q do k1=j+J[k] S[j1]=S[j1]+x[k]*S[k1]
      end for
    end for
  end for 
end if j1=q1 if n[r]>0 then j1=r 
end if if j1=r then

-- To calculate the residuals, the data are read again, in the next
-- block the code numbers are read.

  txt=open("input.txt","r") tx=gets(txt) 
  U=get(txt) U=get(txt) l1=rnd(U[2]) U=get(txt) 
  for i=1 to s do U=get(txt) ds[i]=gets(txt)
  end for for i=1 to t do U=get(txt) U=get(txt) U=get(txt) U=get(txt)
  end for 
end if x[1]=1 z[1]=1 i2=1 for i=1 to j1 do
  if n[i]>0 then
  p=0 nl(2) h=m1[i][s1] t=m1[i][s2] k1=0 if i<r then
    c1=0 for j=1 to s do
      if m1[i][j]=1 then c1=j
      end if k1=k1+abs(m1[i][j])
    end for 
  end if puts(out,tm[i]) if k1=1 then puts(out,"*** ") puts(out,AL[c1])
    puts(out," = ") puts(out,ds[c1]) else nl(1) 
  end if for j=1 to s do puts(out," ") puts(out,AL[j])
  end for puts(out,"     W(I)       MEAN") 
  if ts then puts(out,"  S(MEAN)")
  end if puts(out,"  OBS.MEAN") if i=r then puts(out,"  RESIDU")
  end if y=repeat(0,s) f=repeat(0,s) for j=1 to s do
    if m1[i][j]>0 then p=p+1 f[p]=j
    end if
  end for c=p-1 for j=1 to c do y[f[j]]=1
  end for while TRUE do if i<r and i2>v[i+1] then exit
  end if if i=r then

-- Read data for the next treatment, that is if it is the turn for the
-- residuals to be calculated.

    U=get(txt) if U[1]!=0 then exit 
      end if if dn then
	y[1]=rnd(U[2]) for j=2 to s do U=get(txt) y[j]=U[2]
	end for U=get(txt)
      end if 
    end if if i<r or dn=FALSE then gn()
    end if ML() nl(1) w=0 for j=1 to h do w=w+x[j]*Y[j] 
    end for if ts then
      z0=0 for j=1 to h do
	  o=0 for k=j+1 to h do i1=j+J[k] o=o+x[k]*S[i1]
	end for i1=j+J[j] z0=z0+x[j]*(2*o+x[j]*S[i1])
      end for
    end if
    if z0>0 then z0=sqrt(X*z0) else z0=0
    end if for j=1 to s do printf(out,"%2d",y[j])
    end for if i>=r then puts(out,"          ") 
      else printf(out,"%10.3f",D[1][i2])
      end if printf(out,"%10.3f",w) if ts then printf(out,"%9.3f",z0)
    end if if i<r then printf(out,"%10.3f",B[i2]) i2=i2+1
    else if wt then
	o=U[2] U=get(txt)
      end if if n0 then
      l1=rnd(U[2]) if l1>0 then U=get(txt)
	end if
      end if if l1>0 then
	o=U[2] for j=2 to l1 do
	  U=get(txt) o=o+U[2]
	end for o=o/l1 printf(out,"%10.3f",o) o=o-w printf(out,"%8.3f",o)
      end if
    end if 
  end while
  end if
end for nl(1) close(out)





