-- STAT.E
constant TRUE=1
include math.e

global function PH(atom x) -- right-sided probability in normal distribution.
  atom z,r
  sequence o
  o={0.000488906,0.0000380036,0.0032776263,0.0211410061,0.049867347,1}
  z=abs(x) if z>5 then r=0 else r=0.000005383
  for i=1 to 6 do r=z*r+o[i]
  end for r=power(r,-16)/2
  end if if x>=0 then return r else return 1-r
  end if
end function

global function GM(atom x)  -- gamma-function for x<=3
  atom p,t
  sequence q
  q={-0.0003442342,0.0015397681,-0.002466748,0.0109736958,-0.0002109075,
     0.0742379076,0.0815782188,0.4118402518,0.422784337,1}
  t=x if x>2 then t=x-2 elsif x>1 then t=x-1
  end if p=0.0000677106 for i=1 to 10 do p=t*p+q[i]
  end for if x>0 then if x<=2 then p=p/x 
    end if if x<=1 then p=p/(x+1)
    end if else p=1
  end if return p
end function

global function CH(integer n, atom x) -- Chi-squared distribution
  atom a,c,d,p,t,u,v,w,y,z
  a=n u=2*a u=45*sqrt(u)+a if x>=u then return 0
  end if if a>=50 then u=power(x/a,1/3) v=2/(9*a)
    y=(u-1+v)/sqrt(v) return PH(y)
  end if y=x/2 a=a/2 if a<=3 then
    p=GM(a) w=power(e,a*log(y)-y)/(a*p) else
    u=1/a u=1+u*(1+u*(1-u*(139+u*571/48)/180)/24)/12
    v=2.5066282746*sqrt(a) w=power(e,a*log(y/a)+a-y)/(u*v)
  end if v=w if y>=a then w=w*a/y c=1-w d=(a-1)/y u=d t=1 z=abs(u)
    while z<=1 do if z<1E-7 then return 1-c
      end if w=w*d c=c-w t=t+1 d=(a-t)/y u=u*d z=abs(u)
    end while else
    w=v c=v t=1 u=y/(a+1) d=u z=abs(u)
    while z>1E-7 do w=w*d c=c+w t=t+1 d=y/(a+t) u=power(d,t)
    end while
  end if return 1-c
end function

global function BT(atom x, atom p, atom q) -- Beta-function
  integer k
  atom s,u,v,w
  u=power(x,p) s=u/p k=1
  while TRUE do
    u=(k-q)*x*u/k v=u/(k+p) s=s+v k=k+1 if abs(v)/s<1E-7 then exit
    end if
  end while u=GM(p+q) v=GM(p) w=GM(q) s=s*u/v/w return s
end function

global function FD(atom c, atom d, atom f) -- F-distribution
  integer h,u,r,t
  atom z,i0,i1,x,a,b,p,q,v
  if d>50 and c*d>500 then
    p=2/(9*c) q=2/(9*d) z=power(f,1/3)
    x=((1-q)*z-(1-p))/sqrt(q*z*z+p) return PH(x)
  end if p=d/2 q=c/2 x=d/(d+c*f) if x>.5 then h=1 else h=0
  end if if h then x=1-x p=p-q q=p+q p=q-p
  end if r=floor(p) t=floor(q) p=p-r q=q-t if p<=0 then p=1 r=r-1
  end if z=1-x if q<=0 then q=1 t=t-1
  end if i1=BT(x,p,q) if t>0 then i0=i1 a=p+q-1 b=q i1=BT(x,p,q+1)
    for i=1 to t-1 do z=1-x a=a+1 b=b+1 z=a*z/b
      z=(1+z)*i1-z*i0 i0=i1 i1=z
    end for 
  end if z=i1 i0=z q=q+t u=floor(z*r+1.5)
    while TRUE do
      v=0 i1=z b=u+p a=b+q-1 for i=u to 1 by -1 do
	v=a*x/(a*x+b*(1-v)) a=a-1 b=b-1
	if i<=r then i1=i1*v
	end if
      end for if i1>0 then 
	if abs(i1-i0)/i1<1E-7 then exit
	end if i0=i1 u=u+5
      end if
    end while if h then i1=1-i1
  end if return i1
end function

